from functools import cached_property
from typing import Callable, Optional, Any

from django.core.cache import cache
from django.db import models
from django.db.models import QuerySet, Q
from pydantic import BaseModel

from promy.utils.cache import cache_get_key
from promy.utils.redis import get_redis_client


class QuerysetHashHeap:
    def __init__(
        self,
        *,
        name: str,
        queryset: QuerySet,
        get_node_data: Callable[[Any], str],
    ):
        self.name: str = name
        self.queryset = queryset.order_by("created_at", "pk")
        self.get_node_data = get_node_data

    @cached_property
    def _heap(self) -> "HashHeap":
        heap = HashHeap(
            name=self.name,
            size=lambda: self.queryset.all().count(),
            get_node_data=lambda index: self.get_node_data(
                self.get_index_instance(index)
            ),
        )
        return heap

    def _get_instance_heap_node(self, instance: Any) -> "HashHeapNode":
        return self._heap.get_node(self._instance_index(instance))

    def _instance_index(self, instance: Any) -> int:
        return self.queryset.filter(
            Q(created_at__lt=instance.created_at)
            | Q(created_at=instance.created_at, pk__lte=instance.pk)
        ).count()

    def get_instance_heap_node_dict(self, instance: Any) -> dict | None:
        node = self._get_instance_heap_node(instance)
        return node.dict()

    def get_index_heap_node_dict(self, index: int) -> dict | None:
        return self._heap.get_node(index).dict()

    def get_index_instance(self, index: int) -> Any:
        assert index >= 1
        return self.queryset.all()[index - 1]

    def updated_recently_updated(self, timeout: int):
        r = get_redis_client()
        with r.lock(f"hash_heap:lock:{self.name}", timeout=timeout):
            params = {}

            last_updated_at = cache.get(f"hash_heap:last_updated_at:{self.name}")
            if last_updated_at:
                params["updated_at__gte"] = last_updated_at

            max_updated_at = self.queryset.aggregate(
                max_updated_at=models.Max("updated_at")
            )["max_updated_at"]
            params["updated_at__lte"] = max_updated_at

            for instance in self.queryset.filter(**params).order_by("updated_at"):
                node = self._get_instance_heap_node(instance)
                node.recalculate_with_ancestors(update_data=True)

            cache.set(f"hash_heap:last_updated_at:{self.name}", max_updated_at)

    def rebuild_heap(self, timeout: int):
        r = get_redis_client()
        with r.lock(f"hash_heap:lock:{self.name}", timeout=timeout):
            self._heap.recalculate()


class ReplicatedHashHeap:
    def __init__(
        self,
        *,
        name: str,
        get_node: Callable[[int], dict],
    ):
        self.name: str = name
        self.get_node = get_node

    def _fetch_heap_size(self) -> int:
        return int(self.get_node(0)["size"])

    def _expire_heap_size_cache(self) -> None:
        cache.delete(f"hash_heap:replicated:size:{self.name}")

    def _get_heap_size(self) -> int:
        key = f"hash_heap:replicated:size:{self.name}"
        size = cache.get(key)
        if size is None:
            size = self._fetch_heap_size()
            cache.set(key, size)
        return size

    def _get_heap_node_data(self, index: int) -> str:
        d = self.get_node(index)
        return d["data"]

    @cached_property
    def heap(self) -> "HashHeap":
        heap = HashHeap(
            name=self.name,
            size=self._get_heap_size,
            get_node_data=self._get_heap_node_data,
        )
        return heap

    def _sync_node(self, index: int) -> None:
        reference_node_data = self.get_node(index)
        node = self.heap.get_node(index)

        if node.checksum == reference_node_data["checksum"]:
            return

        if node.left_checksum != reference_node_data["left_checksum"]:
            self._sync_node(index * 2)
        if node.right_checksum != reference_node_data["right_checksum"]:
            self._sync_node(index * 2 + 1)

        node.recalculate(update_data=node.data != reference_node_data["data"])

    def sync(self, timeout: int) -> None:
        r = get_redis_client()
        with r.lock(f"hash_heap:lock:{self.name}", timeout=timeout):
            self._expire_heap_size_cache()
            self._sync_node(1)

    def update_node(self, index: int, timeout: int) -> None:
        r = get_redis_client()
        with r.lock(f"hash_heap:lock:{self.name}", timeout=timeout):
            self.heap.get_node(index).recalculate_with_ancestors(update_data=True)


class HashHeap:
    def __init__(
        self,
        name: str,
        size: int | Callable[[], int],
        get_node_data: Callable[[int], str],
    ):
        self.name = name
        self._size = size
        self.get_node_data = get_node_data

    @property
    def size(self) -> int:
        return self._size() if callable(self._size) else self._size

    def recalculate(self) -> None:
        for index in reversed(range(1, self.size + 1)):
            self.get_node(index).recalculate()

    def get_node(self, index: int) -> "HashHeapNode":
        assert 1 <= index <= self.size
        return HashHeapNode(name=self.name, index=index, heap=self)


class HashHeapNode(BaseModel):
    index: int
    heap: Any
    size: int | None = None
    data: str | None = None
    left_checksum: str | None = None
    right_checksum: str | None = None

    checksum: str | None = None

    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)
        self.reload()

    def _get_parent(self) -> Optional["HashHeapNode"]:
        if self.index == 1:
            return None
        return self.heap.get_node(self.index // 2)

    @property
    def children(self) -> list["HashHeapNode"]:
        children = []
        for idx in (self.index * 2, self.index * 2 + 1):
            if idx < self.heap.size:
                children.append(self.heap.get_node(idx))
        return children

    def child(self, i: int) -> Optional["HashHeapNode"]:
        assert i in (0, 1), "i must be 0 or 1"
        children = self.children
        return children[i] if i < len(children) else None

    @property
    def left(self) -> Optional["HashHeapNode"]:
        return self.child(0)

    @property
    def right(self) -> Optional["HashHeapNode"]:
        return self.child(1)

    def save(self) -> None:
        """
        Save to cache
        """
        cache.set(
            f"heap_node:{self.heap.name}:{cache_get_key(self.index)}", self.dict()
        )

    def reload(self) -> None:
        """
        Re-fetch from cache (and save it to cache)
        """
        data = cache.get(f"heap_node:{self.heap.name}:{cache_get_key(self.index)}")
        if data:
            self.size = data["size"]
            self.checksum = data["checksum"]
        self.save()

    def recalculate(self, update_data: bool = False) -> None:
        self.reload()

        self.size = self.calculate_size()
        if update_data:
            self._update_data()
        self.checksum = self.calculate_checksum()

        self.save()

    def recalculate_with_ancestors(self, update_data: bool = False) -> None:
        self.recalculate(update_data=update_data)
        parent = self._get_parent()
        if parent:
            parent.recalculate_with_ancestors()

    def calculate_size(
        self,
    ) -> int:
        size = 1
        for child in self.children:
            if child.size is None:
                child.recalculate()
            assert child.size is not None
            size += child.size
        return size

    def _update_data(self) -> None:
        self.data = self.heap.get_node_data(self.index)

    def calculate_checksum(self) -> str:
        if not self.data:
            self._update_data()

        left, right = self.left, self.right
        if left:
            left.reload()
        if right:
            right.reload()

        self.left_checksum = left.checksum if left else None
        self.right_checksum = right.checksum if right else None

        if left is not None and self.left_checksum is None:
            left.recalculate()
            self.left_checksum = left.checksum

        if right is not None and self.right_checksum is None:
            right.recalculate()
            self.right_checksum = right.checksum

        return cache_get_key(
            "hash_heap_node_checksum",
            index=self.index,
            size=self.size,
            data=self.data,
            left=self.left_checksum,
            right=self.right_checksum,
        )



from typing import Iterable

from promy.station.registry import station_registry
from promy.utils.celery import app
from promy.utils.monitoring import observe
from promy.station.consts import (
    HEAP_SYNC_TIMEOUT,
    HEAP_UPDATE_TIMEOUT,
    HEAP_REBUILD_TIMEOUT,
)
from promy.station.heap import QuerysetHashHeap, ReplicatedHashHeap


@app.task
@observe("publish_everything", const_labels={"observer_type": "celery_task"})
def publish_everything() -> None:
    for publisher in station_registry.publishers:
        publisher.publish_all()


@app.task
@observe(
    "publish_recently_updated_objects", const_labels={"observer_type": "celery_task"}
)
def publish_recently_updated_objects() -> None:
    for publisher in station_registry.publishers:
        publisher.publish_recently_updated()


@app.task
@observe("update_everything", const_labels={"observer_type": "celery_task"})
def update_everything() -> None:
    if station_registry.periodic_updaters is None:
        return
    station_registry.periodic_updaters.update("all")


@app.task
@observe("update_set", const_labels={"observer_type": "celery_task"})
def update_set(name: str) -> None:
    if station_registry.periodic_updaters is None:
        return
    station_registry.periodic_updaters.update(name)


@app.task
def rebuild_publisher_heaps(timeout: int | None = None) -> None:
    for heap in _get_publisher_heaps():
        heap.rebuild_heap(timeout or HEAP_REBUILD_TIMEOUT)


@app.task
def update_publisher_heaps(timeout: int | None = None) -> None:
    for heap in _get_publisher_heaps():
        heap.updated_recently_updated(timeout or HEAP_UPDATE_TIMEOUT)


def _get_publisher_heaps() -> Iterable[QuerysetHashHeap]:
    for publisher in station_registry.publishers:
        if publisher.enable_heap:
            heap = publisher.heap
            if heap is None:
                continue
            yield heap


@app.task
def sync_updater_heaps(timeout: int | None = None) -> None:
    for updater in _get_updater_heaps():
        updater.sync(timeout or HEAP_SYNC_TIMEOUT)


def _get_updater_heaps() -> Iterable[ReplicatedHashHeap]:
    for updater in station_registry.updater_handlers:
        for heap in updater.get_heaps():
            yield heap



import json
import logging
from abc import abstractmethod
from datetime import datetime, timedelta
from typing import TYPE_CHECKING, TypeVar, Generic, Self, Any, Callable, Iterable
from uuid import UUID

from django.conf import settings
from django.db import transaction
from django.db.models import QuerySet
from django.utils import timezone
from django.utils.timezone import make_aware
from pydantic import BaseModel, ValidationError
from redis.exceptions import LockError
from rest_framework.serializers import SerializerMetaclass

from promy.station.heap import ReplicatedHashHeap
from promy.utils.http_requests import HTTPClient
from promy.utils.models import BaseModel as DjangoBaseModel
from promy.utils.redis import get_redis_client
from promy.utils.cache import cache_get_key
from promy.utils.time import iso_serialize, iso_deserialize

ModelType = TypeVar("ModelType", bound=DjangoBaseModel)
BoundedType = TypeVar("BoundedType", bound=DjangoBaseModel)

logger = logging.getLogger(__name__)

if TYPE_CHECKING:
    from django.db.models.manager import RelatedManager  # type: ignore

    RelatedManagerType = RelatedManager[ModelType]
else:
    RelatedManagerType = Any


class SkipUpdate(Exception):
    pass


class Updater:
    @abstractmethod
    def update(self):
        ...


class IterableUpdater(Updater):
    def __init__(self, col: Iterable[Updater]):
        self.col = col

    def update(self):
        for updater in self.col:
            updater.update()


class QuerysetUpdater(Updater):
    def __init__(self, qs: QuerySet):
        self.qs = qs

    def update(self):
        for updater in self.qs.all():
            updater.update()


class CacheUpdater(Updater):
    def __init__(
        self,
        queryset: QuerySet | Callable,
        serializer: SerializerMetaclass | Callable,
        cache_key_prefix: str,
        periodic_should_update_func: Callable | None = None,
        get_value_should_update_func: Callable | None = None,
        lock_timeout_seconds: int = 300,
    ):
        self.queryset = queryset
        self.serializer = serializer
        self.cache_key_prefix = cache_key_prefix
        if periodic_should_update_func is not None:
            self.periodic_should_update_func = periodic_should_update_func
        if get_value_should_update_func is not None:
            self.get_value_should_update_func = get_value_should_update_func
        self.lock_timeout_seconds = lock_timeout_seconds
        self.batch_size = 1000
        self.redis = get_redis_client()
        self.now = datetime.now()

    def update(self):
        try:
            with self.redis.lock(
                self.get_cache_key(), self.lock_timeout_seconds, blocking=False
            ):
                self._update()
        except LockError:
            ...

    def _update(self):
        self.now = datetime.now()
        start_index = 0
        if isinstance(self.queryset, QuerySet):
            all_objects = self.queryset.all()
            count = self.queryset.count()
        else:
            all_objects = self.queryset()
            count = len(all_objects)
        while start_index < count:
            query_set = all_objects[start_index : start_index + self.batch_size]
            start_index += self.batch_size
            last_update_keys = [
                self.get_last_update_timestamp_redis_key(instance)
                for instance in query_set
            ]
            last_update_timestamps = self.redis.mget(last_update_keys)
            for instance, last_update_timestamp in zip(
                query_set, last_update_timestamps
            ):
                try:
                    last_update = None
                    if last_update_timestamp is not None:
                        last_update = make_aware(
                            datetime.fromtimestamp(float(last_update_timestamp))
                        )  # type: ignore
                    if self.periodic_should_update(instance, last_update):
                        self.force_update(instance)
                except Exception:
                    logger.exception(f"could not cache {instance}")

    def periodic_should_update(
        self, instance, last_update: datetime | None = None
    ) -> bool:
        if last_update is None:
            last_update = self.get_last_update(instance)
        if self.periodic_should_update_func is None:
            return self.default_should_update(instance, last_update)
        return self.periodic_should_update_func(instance, last_update)

    def get_value_should_update(self, instance) -> bool:
        last_update = self.get_last_update(instance)
        if self.get_value_should_update_func is None:
            return self.default_should_update(instance, last_update)
        return self.get_value_should_update_func(instance, last_update)

    def default_should_update(self, instance, last_update: datetime | None) -> bool:
        return last_update is None or last_update < instance.updated_at

    def get_last_update(self, instance) -> datetime | None:
        last_update_timestamp = self.redis.get(
            self.get_last_update_timestamp_redis_key(instance)
        )
        if last_update_timestamp is None:
            return None
        return make_aware(
            datetime.fromtimestamp(float(last_update_timestamp))
        )  # type: ignore

    def get_value(self, instance):
        if self.get_value_should_update(instance):
            value = self.force_update(instance)
        else:
            value = self.redis.get(self.get_value_redis_key(instance))
            if value is None:
                value = self.force_update(instance)
        return json.loads(value)

    def force_update(self, instance) -> str:
        if isinstance(self.serializer, SerializerMetaclass):
            value = json.dumps(self.serializer(instance).data)
        else:
            value = json.dumps(self.serializer(instance))
        self.redis.set(self.get_value_redis_key(instance), value)
        self.redis.set(
            self.get_last_update_timestamp_redis_key(instance), self.now.timestamp()
        )
        return value

    def get_key(self, instance):
        if hasattr(instance, "key"):
            return instance.key
        return cache_get_key(instance)

    def get_last_update_timestamp_redis_key(self, instance) -> str:
        return f"{self.cache_key_prefix}:{self.get_key(instance)}:timestamp"

    def get_value_redis_key(self, instance) -> str:
        return f"{self.cache_key_prefix}:{self.get_key(instance)}:value"

    def get_cache_key(self) -> str:
        return f"{self.cache_key_prefix}:run_lock"


class ModelUpdater(BaseModel, Generic[ModelType], Updater):
    @abstractmethod
    def update(self) -> ModelType:
        ...


class BindingUpdater(BaseModel, Generic[ModelType, BoundedType]):
    @abstractmethod
    def update(self, bounded: BoundedType) -> ModelType:
        ...

    @classmethod
    def update_bindings(
        cls,
        *,
        updaters: list[Self],
        bounded: BoundedType,
        bindings: RelatedManagerType,  # type: ignore
    ) -> list[ModelType]:
        with transaction.atomic():
            instances = [u.update(bounded) for u in updaters]
            bindings.exclude(key__in=[i.key for i in instances]).delete()  # type: ignore
            return instances


class UpdaterHandler:
    def __init__(
        self,
        default_object_name: str | None = None,
        client: HTTPClient | None = None,
    ) -> None:
        self.handlers: dict[
            str, tuple[type[Updater], tuple[Callable[[Any], None], ...]]
        ] = {}
        self.default_object_name = default_object_name

        self.client = client
        if self.client:
            assert self.client.base_url, "base_url is not provided to HTTPClient"

    def add(
        self,
        object_name: str,
        updater_class: type[Updater],
        *callbacks: Callable[[Any], None],
    ) -> Self:
        assert object_name not in self.handlers
        self.handlers[object_name] = updater_class, callbacks
        return self

    def pull(self, name: str, **kwargs) -> list[Any]:
        assert self.client, "client is not provided to UpdaterHandler"
        resp = self.client.get(
            "station/publisher/query",
            params={"name": name, **kwargs},
        )
        resp.raise_for_status()
        resp_body = resp.json()
        results = resp_body["results"]
        instances = []

        for item in results:
            instance = self(item, default_name=name)
            if instance:
                instances.append(instance)

        return instances

    def pull_one(self, name: str, **kwargs) -> Any | None:
        instances = self.pull(name, **kwargs)
        if not instances:
            return None

        assert len(instances) == 1
        return instances[0]

    def get_heaps(self) -> Iterable[ReplicatedHashHeap]:
        for name, (updater_class, _) in self.handlers.items():
            heap = self._get_heap(name, updater_class)
            if heap:
                yield heap

    def _get_heap(
        self, name: str, updater_class: type[Updater]
    ) -> ReplicatedHashHeap | None:
        if self.client and getattr(updater_class, "heap_enabled", None):
            return ReplicatedHashHeap(
                name=name,
                get_node=lambda index: self._fetch_heap_node(name, index),
            )
        return None

    def _fetch_heap_node(self, name: str, index: int) -> dict:
        assert self.client
        resp = self.client.get(
            "station/heap/node",
            params={"name": name, "index": str(index)},
        )
        resp.raise_for_status()
        d = resp.json()
        self(d, update_heap=False, raise_exception=True)
        return d["heap_node"]

    def __call__(
        self,
        value,
        *,
        update_heap: bool = True,
        raise_exception: bool = False,
        default_name: str | None = None,
    ) -> Any | None:
        heap_node = None

        if "object_name" in value:
            name = value["object_name"]
            data = value["body"]
            heap_node = value.get("heap_node")

        elif default_name or self.default_object_name:
            name = default_name or self.default_object_name
            data = value
        else:
            logger.info("unknown object received")
            return None

        if name not in self.handlers:
            logger.info("unknown object received", extra={"object_name": name})
            return None

        updater_class, callbacks = self.handlers[name]

        def do_heap_update():
            if not heap_node or not update_heap:
                return
            heap = self._get_heap(heap_node["name"], updater_class)
            heap.update_node(int(heap_node["index"]), timeout=10)

        try:
            data["updater_handler"] = self
            updater = updater_class(**data)
        except ValidationError as e:
            if raise_exception:
                raise e
            logger.exception("invalid object received")
            return None

        try:
            instance = updater.update()
        except SkipUpdate:
            logger.info("update skipped")
            do_heap_update()
            return None

        if instance is None:
            do_heap_update()
            return None

        if isinstance(instance, DjangoBaseModel):
            logger.info(
                "object persisted",
                extra={
                    "model": instance.__class__.__name__,
                    "key": instance.key,
                    "created_at": instance.created_at,
                    "updated_at": instance.updated_at,
                },
            )
        else:
            logger.info("object persisted", extra={"object_name": name})

        for callback in callbacks:
            try:
                callback(instance)
            except Exception as e:
                if raise_exception:
                    raise e
                logger.exception("callback failed")

        do_heap_update()
        return instance


class UpdaterSet:
    def __init__(self, updaters: dict[str, Updater]) -> None:
        self.updaters = updaters

    def add(self, name: str, updater: Updater) -> None:
        self.updaters[name] = updater

    def update(self, name: str | None = None) -> None:
        if name:
            self.updaters[name].update()
        else:
            for updater in self.updaters.values():
                updater.update()


class SpreadUpdater(Updater):
    """
    SpreadUpdater is used to spread the update of a queryset over time (`period`).
    Each element of the queryset is updated once every `period`.
    """

    def __init__(
        self,
        name: str,
        queryset: QuerySet,
        update_func: Callable,
        period: timedelta,
    ) -> None:
        self.name = name
        self.queue_name = f"{SpreadUpdater.redis_prefix()}:queue:{name}"
        self.queryset = queryset
        self.func = update_func
        self.period = period

    def _et_key(self, k: str | UUID) -> str:
        return f"{self.redis_prefix()}:et:{k}"

    @staticmethod
    def redis_prefix() -> str:
        return f"{settings.ENV.service_name}:promy:spread_updater:"

    def update(self) -> None:
        now = timezone.now()

        r = get_redis_client()
        next_task_arr = r.lrange(self.queue_name, -1, -1)

        if next_task_arr:
            elem_key = next_task_arr[0].decode()  # type: ignore
            elem_time = iso_deserialize(r.get(self._et_key(elem_key)).decode())

            if elem_time <= now:
                elem = self.queryset.get(key=elem_key)
                self.func(elem)

                # pop queue
                if e_raw := r.rpop(self.queue_name):
                    e = e_raw.decode()  # type: ignore
                    if e != elem_key:
                        logger.error(
                            f"Schedule updater {self.name}: expected {elem_key} as head of redis queue, got {e}"
                        )
                        r.lpush(self.queue_name, e)

        else:
            count = self.queryset.all().count()
            if count == 0:
                return

            step = self.period / count
            t = now
            for elem in self.queryset.all():
                t += step
                key = str(elem.key)
                r.set(self._et_key(key), iso_serialize(t))
                r.lpush(self.queue_name, key)
